from __future__ import absolute_import, division, print_function
import os
import numpy as np
import tensorflow as tf
import logging
from PIL import Image

from ops import *

# Change logger Level to add more verbose output
logger = logging.getLogger('infer')
logger.setLevel(logging.WARNING)

formatter = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s: %(message)s')
handler = logging.StreamHandler()
handler.setFormatter(formatter)

logger.addHandler(handler)


# Class containing network graph definition
class Branch:

    def __init__(self, dims, h, w, gpu_id=None):
        if type(h) is not int:
            h = int(h)

        if type(w) is not int:
            w = int(w)

        # Set device
        device = '/device:GPU:{}'.format(gpu_id) if gpu_id is not None else '/cpu:0'
        logger.info('Created network in device "{}"'.format(device))

        self.layers = [None] * 16 # Net has 16 layers in total
        self.img_shape = [h, w, 3]
        self.x = tf.placeholder(dtype=tf.float32, shape=[None] + self.img_shape)

        with tf.device(device):
            with tf.variable_scope('vars'):
                # Convolution + pooling + norm 1
                self.layers[0] = conv2d(self.x, k_size=11, n_filters=96, strides=4, relu=True, name='conv1')
                self.layers[1] = max_pool(self.layers[0], name='pool1')
                self.layers[2] = lrn(self.layers[1], name='lrn1')

                # Convolution + pooling + norm 2
                self.layers[3] = conv2d(self.layers[2], 5, 256, 1, True, 'conv2')
                self.layers[4] = max_pool(self.layers[3], name='pool2')
                self.layers[5] = lrn(self.layers[4], name='lrn2')

                # Convolutions 3 & 4
                self.layers[6] = conv2d(self.layers[5], 3, 384, 1, True, 'conv3')
                self.layers[7] = conv2d(self.layers[6], 3, 384, 1, True, 'conv4')

                # Convolutions + pooling 5
                self.layers[8] = conv2d(self.layers[7], 3, 256, 1, True, 'conv5')
                self.layers[9] = max_pool(self.layers[8], name='pool5')

                # "Dense" with dropout 6
                self.layers[10] = conv2d(self.layers[9], 6, 4096, 1, True, 'fc6')
                self.layers[11] = dropout(self.layers[10], 1.0, name='drop6')

                # "Dense" with dropout 7
                self.layers[12] = conv2d(self.layers[11], 1, 4096, 1, True, 'fc7')
                self.layers[13] = dropout(self.layers[12], 1.0, name='drop7')

                # Low res descriptor
                self.layers[14] = conv2d(self.layers[13], 1, 32, 1, False, 'score')

                # Deconvolution (bilinear upsampling + convolution)
                self.layers[15] = upconv2d(self.layers[14], self.x.get_shape()[1:3], 11, dims, 1, 'upconv1')


def load(model_dir, dims, img_shape, gpu_id=None):
    # Reset graph every time
    tf.reset_default_graph()

    if type(dims) is not int:
        dims = int(dims)

    # Set network parameters
    h, w = img_shape
    model_path = os.path.join(model_dir, 'model_{}D'.format(dims))
    logger.info('Loading network from "{}"...'.format(model_path))
    logger.debug('Dims: {} - Height: {} - Width: {} - GPU_ID: {}'.format(dims, h, w, gpu_id))

    # Create network
    net = Branch(dims=dims, h=h, w=w, gpu_id=gpu_id)

    # Init session & load model
    sess = tf.Session()
    tf.train.Saver().restore(sess, model_path)
    logger.info('Network loaded...')

    return [sess, net]


def run(sess, net, img, out_layers):
    # Deal with ugly Matlab conversions
    if type(sess) is list:
        sess = sess[0]

    if type(net) is list:
        net = net[0]

    if type(out_layers) is float:
        out_layers = [int(out_layers)]

    if type(out_layers) is int:
        out_layers = [out_layers]

    if type(out_layers) is not list:
        out_layers = [int(i) for i in out_layers]

    # Parse ugly Matalb converted image
    x = _parse_image(img, net.img_shape)

    logger.info('Running infer on parsed image...')
    out = sess.run([net.layers[i] for i in out_layers], feed_dict={net.x: x})
    logger.info('Infer completed...')

    # Remove 0th dim if batch size is 1
    return [np.squeeze(o) for o in out]


def close(sess):
    if type(sess) is list:
        sess = sess[0]
    sess.close()
    logger.info('Session closed...')


def _parse_image(img, target_shape):

    logger.debug('Original image shape: {}'.format(img.shape))
    img = img.astype(np.float32) / 255.0

    if img.ndim < 2:
        # RGB or Greyscale
        h, w = target_shape[:-1]
        c = img.size // (h*w)

        # Reshape to correct size
        img = np.reshape(img, [c, w, h])
        img = np.transpose(img, axes=[2, 1, 0])

    # Add 4th dimension
    if img.ndim < 4:
        img = np.expand_dims(img, axis=0)

    # Repeat grayscale image over RGB channels
    if img.shape[-1] != 3:
        img = np.repeat(img, 3, axis=3)

    logger.debug('Parsed image shape: {}'.format(img.shape))
    return img


def main():
    """MWE"""
    # Default params
    model_dir = 'fill_model_dir'
    dims = 3
    gpu_id = None

    # Load network & session
    sess, net = load(model_dir, dims, [1080, 1920], gpu_id)

    # Load image
    img = np.array(Image.open('fill_img_path'), dtype=np.float32) / 255.0

    # Generate descriptor
    print(run(sess, net, img, [1, 2]))

    # Close session
    close(sess)


if __name__ == "__main__":
    main()

